library pgraph,nlsys;
@#include c:\andre\gauss\mylibs\probs.src;@

_glb_nsim = 1000;               @ number of simulations                      @
_glbT = 100;                    @ the length of the time series              @
_glbk = 2;                      @ the dimension of the process               @
_glb_psim = 500;                @ number of simulations for asymptotic
                                  p-values                                   @
_glb_fname = "mixnorm.out";     @ file name to write simulations to; header
                                  of file must be filled in below            @
_glbr = 0;                      @ the postulated cointegrating rank          @
_glbp = 1;                      @ the order of the VAR model                 @
_glb_detpol = 0;                @ degree of deterministic polynomial in
                                  VAR model (<0 for no deterministics)       @
_glb_lownu = 3;                 @ lower bound for degrees of freedom of
                                  Student t (>200 for Gaussian)              @
_glb_mett = 1;                  @ if /= 0, also estimate degrees of freedom  @
_glbp_pol = 0;                  @ order of the SNP expansion                 @
_glb_true_degree = 1;           @ if 1, then _glbp_pol is the true degree of
                                  the polynomial as opposed to having some
                                  restricted polynomial of degree
                                  _glbp_pol^_glbk                            @
_glb_usevar = 0;                @ going to be a vector indicating which
                                  parameters to optimize over                @
_glb_johansen = 0;              @ global indicating last trace test computed @
_glb_showcml = 0;               @ __output option for cml                    @
ccs = {0,20};                   @ local stationarity parameters              @
pp_max = 3;                     @ maximum degree of polynomial in sim's      @
without_drift=1;                @ true DGP without drift (1=yes)             @


/****************************
 ** Use maximum likelihood **
 ** procedures of maxlik() **
 ****************************/
library cml,nlsys@,probs,pgraph@; time0=date; format /rd 8,4;
blank=chrs(32+zeros(1,70)); low_nu_org=_glb_lownu;
trap 1;

/*****************************
 ** Loop over distributions **
 *****************************/
distributions={9};      @ innovations distributions, see below; ANDRE @

dist_tel=1; do while dist_tel<=rows(distributions);
dist=distributions[dist_tel];

/***********************
 ** Reset output file **
 ***********************/
if dist==0;
    _glb_fname="gaussian.out";
elseif dist==1;
    _glb_fname="t3.out";
elseif dist==2;
    _glb_fname="t2.out";
elseif dist==3;
    _glb_fname="censcauc.out";
elseif dist==4;
    _glb_fname="chi3.out";
elseif dist==5;
    _glb_fname="F3_3.out";
elseif dist==6;
    _glb_fname="gmix.out";
elseif dist==7;
    _glb_fname="mixnorm2.out";
elseif dist==8;
    _glb_fname="polynorm.out";
elseif dist==9;
    _glb_fname="tmix.out";
else;
    _glb_fname="gaussian.out";
endif;
@output file=^_glb_fname reset;@
output file=^_glb_fname on;
output off;

/****************************
 ** Loop over alternatives **
 ****************************/
ctel=1; do while ctel<=rows(ccs); cc=ccs[ctel];

    /******************************************************
     ** Open file, write header, and perform simulations **
     ******************************************************/
    output on;
    _glb_fname " distribution (alpha restart), cc = " cc ", nu = " _glb_lownu;
    "T=" _glbT ";  dim=" _glbk ";  #sims=" _glb_nsim;;
    ";  VAR(" _glbp ");  ";
    "pval_sim=" _glb_psim ";  det=" _glb_detpol;;
    ";  SNP(" _glbp_pol ");  r=" _glbr;
    output off;

    dustbin=0; iter=1; do while iter<=_glb_nsim;

        /*****************************************************
         ** Generate data and construct regression matrices **
         *****************************************************/
        if dist==0;
            strt=rndn(_glbT,_glbk);
        elseif dist==1;
            strt=rndn(_glbT,_glbk)./sqrt(meanc(rndn(3,_glbT).^2));
        elseif dist==2;
            strt=rndn(_glbT,_glbk)./sqrt(meanc(rndn(2,_glbT).^2));
        elseif dist==3;
            STOP;
            @ Truncated Cauchy @
            strt={}; do while (rows(strt)<_glbT);
                x=_glbT-rows(strt); x=rndn(x,_glbk)./rndn(x,1);
@                x=selif(x,(sumc((x.*x)')/_glbk).<=invcdff(0.95,_glbk,1));@
                if ismiss(x)/=1; strt=strt|x; endif;
            endo;
        elseif dist==4;
            STOP;
@            strt=3; strt=rndchi(_glbT,_glbk,strt)-strt;@
        elseif dist==5;
            STOP;
@            strt=3; strt=rndf(_glbT,_glbk,strt,strt)-strt/(strt-2);@
        elseif dist==6;
            gdist={}; do while cols(gdist) < _glbk;
                strt=rndu(_glbT,1); strt=1+(strt.>0.5)+(strt.>0.8);
                data={0 -1.5,1.5 1.1666666667,-2.25 2};
                gdist=gdist~(rndn(_glbT,2)+data[strt,.]);
            endo;
            strt=gdist[.,1:_glbk];
        elseif dist==7;
            STOP;
            strt=rndu(_glbT,1); strt=1+(strt.>1/3)+(strt.>2/3);
            data=(3~sqrt(3))|(-3~sqrt(3))|(0~-2*sqrt(3));
            strt=rndn(_glbT,_glbk)+data[strt,.];
        elseif dist==8;
            STOP;
            strt={}; do while rows(strt)<_glbT;
                x=rndu(1,1); y=2.2*((x>0.5) - (x<0.5)); num_el=x; __output=0;
                {x,y,data,num_el}=nlsys(&polynorm,y);
                if num_el == 1; strt=strt|x; endif;
            endo;
            if _glbk>1; strt=rndn(_glbT,_glbk-1)~strt; endif;
        elseif dist==9;
            gdist={}; do while cols(gdist) < _glbk;
                strt=rndu(_glbT,1); strt=1+(strt.>0.5)+(strt.>0.8);
                data={0 -1.5,1.5 1.1666666667,-2.25 2};
                gdist=gdist ~ (rndn(_glbT,2)./sqrt(meanc(rndn(3,_glbT).^2)) +
                     data[strt,.]);
            endo;
            strt=gdist[.,1:_glbk];
        else;
            strt=rndn(_glbT,_glbk);
        endif;
        data=zeros(1,_glbk)|cumsumc((without_drift==0)+strt);
        data[.,1]=recserar((0|strt[.,1]),0,1-cc/_glbT);
        {y,x}=gen_xy(data,_glbp,_glb_detpol);

        /**********************************
         ** Loop over SNP approximations **
         **********************************/
        num_el=3; aid0=zeros(1,(pp_max+1)*num_el); par={};
        _glbp_pol=0; tmp_dustbin=0; do while _glbp_pol<=pp_max;

            /*****************************
             ** Compute starting values **
             ** without a22             **
             *****************************/
            _glb_lownu=low_nu_org+2*_glbk*_glbp_pol*(_glb_mett==1);
            strt=comp_strt(y,x,0,par,_glbp_pol-1);

            /******************************************
             ** Compute maximum likelihood estimates **
             ******************************************/
            _cml_MaxIters=400; _cml_CovPar=3; _cml_Algorithm=1;
            _clm_GradMethod=1; __output=_glb_showcml;
            {par,f,g,v,r}=cml(y~x,seqa(1,1,cols(y)+cols(x)),
                                 &comp_loglik,strt);
            if r == 2;
                _cml_MaxIters=200; _cml_CovPar=3; _cml_Algorithm=3;
                {par,f,g,v,r}=cml(y~x,seqa(1,1,cols(y)+cols(x)),
                                     &comp_loglik,par);
            endif;
            __output=2; tmp_dustbin=tmp_dustbin+(r/=0);
            if r /= 0; break; endif;
            aid0[_glbp_pol*num_el+1]=f;
            if iter+_glbp_pol > 1;
                cls; locate 2+iter%20,1; blank; blank; blank;
                locate 2+iter%20,1;
                "Iteration " iter " of " _glb_nsim;;
                " (" ftos(ctel,"%*.*lf",1,0) ":";;
                ftos(rows(ccs),"%*.*lf",1,0) ";    ";;
                ftos(_glbp_pol,"%*.*lf",1,0) ":";;
                ftos(1+pp_max,"%*.*lf",1,0) ")";
                "Dustbin: " dustbin " iterations";
                "Elapsed : " etstr(ethsec(time0,date));
                lm=(dist_tel-1)*rows(ccs)*_glb_nsim*(1+pp_max)+
                   (ctel-1)*_glb_nsim*(1+pp_max)+
                   (iter-1)*(1+pp_max)+
                   _glbp_pol;
                gr=etstr(round(ethsec(time0,date)/lm));
                "Average Time per Simulation : " gr;
                "Estimated Total Remaining Time : ";;
                etstr(round(ethsec(time0,date)/lm*(rows(distributions)*
                      rows(ccs)*(1+pp_max)*_glb_nsim-lm)));
                "Output to file " _glb_fname;
            endif;

            /********************************************
             ** Compute gradient, Hessian, and LM test **
             ********************************************/
            {gr,he,op,lm}=comp_grheop(par,(y~x),_glbk,_glbr);
            {c1,c2}=comp_corr(par,(y~x));
            {pval}=sim_pval(rows(y),lm,c1,c2);
            aid0[_glbp_pol*num_el+2]=pval[1+without_drift];
            aid0[_glbp_pol*num_el+3]=rows(par);

        _glbp_pol=_glbp_pol+1; endo;

        if tmp_dustbin /= 0;
            dustbin=dustbin+1; iter=iter-1;
        else;
            @ compute Johansen test, Gaussian LM and t(5) LM tests @
            if _glb_mett/=0;
                _glb_mett=0; _glb_lownu=500; _glbp_pol=0;
                _clm_GradMethod=1; __output=_glb_showcml;
                par={}; strt=comp_strt(y,x,0,par,_glbp_pol-1);
                {par,f,g,v,r}=cml(y~x,seqa(1,1,cols(y)+cols(x)),
                                     &comp_loglik,strt);
                {gr,he,op,lm}=comp_grheop(par,(y~x),_glbk,_glbr);
                {c1,c2}=comp_corr(par,(y~x)); {pval}=sim_pval(rows(y),lm,c1,c2);
                aid0=aid0~f~pval[1+without_drift]~rows(par)~_glb_johansen;

                _glb_lownu=5; par={}; strt=comp_strt(y,x,0,par,_glbp_pol-1);
                {par,f,g,v,r}=cml(y~x,seqa(1,1,cols(y)+cols(x)),
                                     &comp_loglik,strt);
                _glb_lownu=low_nu_org;
                if r/=0;
                    dustbin=dustbin+1; iter=iter-1;
                else;
                    {gr,he,op,lm}=comp_grheop(par,(y~x),_glbk,_glbr);
                    {c1,c2}=comp_corr(par,(y~x));
                    {pval}=sim_pval(rows(y),lm,c1,c2);
                    aid0=aid0~f~pval[1+without_drift]~rows(par);
                    _glb_mett=1;

                    @ print everything @
                    screen off; output on; aid0; output off; screen on;
                endif;

            else;
                @ print everything @
                screen off; output on; aid0; output off; screen on;
            endif;

        endif;

    iter=iter+1; endo;

    output on; "Simulations in dustbin: " dustbin; ""; ""; ""; ""; output off;

ctel=ctel+1; endo;

output on; cls; "Elapsed Time : " etstr(ethsec(time0,date)); output off;

dist_tel=dist_tel+1; endo;

/*****************************************************************************
 *****************************************************************************
 ****                                                                     ****
 ****                                                                     ****
 ****     BBBB  EEEEE LL     OOO  WW     WW       AAAA  RRRR    EEEEE     ****
 ****     BB  B EE    LL    OO OO WW     WW      AA  AA RR RR   EE        ****
 ****     BBBB  EEEE  LL    OO OO WW  W  WW      AAAAAA RRRR    EEEE      ****
 ****     BB  B EE    LL    OO OO WW WWW WW      AA  AA RR RR   EE        ****
 ****     BBBB  EEEEE LLLLL  OOO   WWW WWW       AA  AA RR  RR  EEEEE     ****
 ****                                                                     ****
 ****                                                                     ****
 ****  TTTTTT HH  HH  EEEEE     SSSS EEEEE PPPP   AAAA  RRRR    AAAA      ****
 ****    TT   HH  HH  EE       SS    EE    PP PP AA  AA RR RR  AA  AA     ****
 ****    TT   HHHHHH  EEEE      SSS  EEEE  PPPP  AAAAAA RRRR   AAAAAA --  ****
 ****    TT   HH  HH  EE          SS EE    PP    AA  AA RR RR  AA  AA     ****
 ****    TT   HH  HH  EEEEE    SSSS  EEEEE PP    AA  AA RR  RR AA  AA     ****
 ****                                                                     ****
 ****                                                                     ****
 ****  TTTTTT EEEEE     RRRR    OOO  UU UU TTTTTT II NNN   N EEEEE   SSSS ****
 ****    TT   EE        RR RR  OO OO UU UU   TT   II NNNN  N EE     SS    ****
 ****    TT   EEEE      RRRR   OO OO UU UU   TT   II NN NN N EEEE    SSS  ****
 ****    TT   EE        RR RR  OO OO UU UU   TT   II NN  NNN EE        SS ****
 ****    TT   EEEEE     RR  RR  OOO   UUU U  TT   II NN   NN EEEEE  SSSS  ****
 ****                                                                     ****
 ****                                                                     ****
 *****************************************************************************
 *****************************************************************************
 *****************************************************************************
 *****************************************************************************/




proc (1)=p_air(nrs,base,digits);
/***************************************************************************
 ***************************************************************************
 **                                                                       **
 ** This routine converts the (n x 1) vector nrs to a (n x digits) matrix **
 ** containing the representation of nrs in a system with base "base" and **
 ** with digits "digits". For example,                                    **
 **                                                                       **
 ** p_air(9|114,2,8)={0 0 0 0 1 0 0 1,0 1 1 1 0 0 1 0}                    **
 **                                                                       **
 ** Author: Andre Lucas, January 17, 1997.                                **
 **                                                                       **
 ***************************************************************************
 ***************************************************************************/

    local sgn,result,i1;

    if cols(nrs)/=1;
        "*** P_AIR() ONLY IMPLEMENTED FOR SINGLE COLUMNS ***";
        stop;
    endif;

    sgn=(nrs.>0)-(nrs.<0); result={}; nrs=abs(nrs); i1=1; do while i1<=digits;
        result=(nrs%base)~result; nrs=(nrs-result[.,1])/base;
    i1=i1+1; endo;

    if sumc(nrs)/=0;
        "*** INSUFFICIENT DIGITS TO CONVERT NUMBERS IN P_AIR() ***";
        stop;
    endif;

    retp(sgn.*result);
endp;






proc comp_C(alpha,p,n,nu,mom);
/******************************************************************************
 ******************************************************************************
 **                                                                          **
 ** This routine computes the normalizing constant of the SNP density kernel **
 **                                                                          **
 ** INPUT                                                                    **
 ** alpha: the coeficients of the polynomial in the SNP expansion around the **
 **        Student t. alpha is parameterized as follows. Let A=alpha[i+1],   **
 **        then A is the coefficient of x[1]^k[1]*...*x[n]^k[n], where       **
 **        k[1]...k[n] is the n-digit representation of i in a system with   **
 **        base p.                                                           **
 ** p    : the degree of the polynomial in the Gallant-Nytchka               **
 **        representation. Note that this implies a restricted polynomial    **
 **        of degree p^n.                                                    **
 ** n    : the dimension of the stochastic process                           **
 ** nu   : the degrees of freedom of the Student t distribution, the         **
 **        central part of the SNP expansion. nu>=200 gives the normal       **
 **        distribution. Note that nu must be large enough for the           **
 **        appropriate moments to exist due to the polynomial tails of       **
 **        the Student t and the polynomial in front of the central          **
 **        density in the SNP expansion.                                     **
 ** mom  : moment to compute (0 for integrating constant, 1 for mean, etc.)  **
 **                                                                          **
 ** OUTPUT                                                                   **
 ** C    : the normalizing constant of the SNP density.                      **
 **                                                                          **
 ** Author: Andre Lucas, January 17, 1997.                                   **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/

    local c0,C,aid1,aid2,aid3,aid4,aid5,aid6,i1;

    /***********************
     ** Compute c0(1..2p) **
     ***********************/
    c0=cumprodc(1|seqa(1,2,maxc(1|p)+1)); c0=c0.*.(1|0); c0=c0[1:2*p+2];

    /*********************************************************************
     ** Compute normalizing constant in steps. 1st compute the separate **
     ** terms, then sum.                                                **
     *********************************************************************/
    C=alpha.*.alpha;    @ 1st part is product of alphas @
    aid3=(C./=0);       @ only those moments need exist for which alpha/=0 @
    aid1=p_air(seqa(0,1,rows(alpha)),p+1,n);
                        @ compute power indices for different elements @
    if mom==0;
        aid2=1+((ones(rows(alpha),1).*.aid1) + (aid1.*.ones(rows(alpha),1)));
        C=C.*prodc(reshape(c0[aid2],rows(alpha)^2,n)');
                        @ multiply by appropriate normal moments @
        aid2=sumc(aid1');
        aid2=((ones(rows(alpha),1).*.aid2) + (aid2.*.ones(rows(alpha),1)));
        aid2=(aid3.*aid2);
        if nu<200;          @ multiply by Student-t factor @
            C=C.*exp(aid2*ln(nu/2)/2 - lnfact(nu/2-1) + lnfact((nu-aid2)/2-1));
        endif;
        C=sumc(C);
    elseif mom > 0;
        aid5={}; i1=1; do while i1<=n;
            aid4=aid1; aid4[.,i1]=aid4[.,i1]+mom;
            aid2=1+((ones(rows(alpha),1).*.aid1) +
                 (aid4.*.ones(rows(alpha),1)));
            aid6=C.*prodc(reshape(c0[aid2],rows(alpha)^2,n)');
                            @ multiply by appropriate normal moments @
            aid2=sumc(aid1');
            aid2=((ones(rows(alpha),1).*.aid2) + (aid2.*.ones(rows(alpha),1)));
            aid2=aid2+1; aid2=(aid3.*aid2);
            if nu<200;          @ multiply by Student-t factor @
                aid6=aid6.*exp(aid2*ln(nu/2)/2 -
                        lnfact(nu/2-1) + lnfact((nu-aid2)/2-1));
            endif;
            aid5=aid5|sumc(aid6);
        i1=i1+1; endo;
        C=aid5;
    else;
        "*** NOT IMPLEMENTED FOR NEGATIVE MOMENTS : COMP_C() ***";
        stop;
    endif;

    retp(C);
endp;






proc (1)=comp_logkernel(x,alpha,p,nu);
/******************************************************************************
 ******************************************************************************
 **                                                                          **
 ** This routine computes ln() of the kernel of the SNP density, so without  **
 ** the normalizing constant.                                                **
 **                                                                          **
 ** INPUT                                                                    **
 ** x    : the data, a (T x n) matrix. By the data is meant the stochastic   **
 **        process that is to be approximated by the (orthogonalized) SNP    **
 **        expansion, so usually the scaled error process of the regression. **
 ** alpha: the coeficients of the polynomial in the SNP expansion around the **
 **        Student t. alpha is parameterized as follows. Let A=alpha[i+1],   **
 **        then A is the coefficient of x[1]^k[1]*...*x[n]^k[n], where       **
 **        k[1]...k[n] is the n-digit representation of i in a system with   **
 **        base p.                                                           **
 ** p    : the degree of the polynomial in the Gallant-Nytchka               **
 **        representation. Note that this implies a restricted polynomial    **
 **        of degree p^n.                                                    **
 ** nu   : the degrees of freedom of the Student t distribution, the         **
 **        central part of the SNP expansion. nu>=200 gives the normal       **
 **        distribution. Note that nu must be large enough for the           **
 **        appropriate moments to exist due to the polynomial tails of       **
 **        the Student t and the polynomial in front of the central          **
 **        density in the SNP expansion.                                     **
 **                                                                          **
 ** OUTPUT                                                                   **
 ** kern : the ln(kernel) of the SNP density.                                **
 **                                                                          **
 ** Author: Andre Lucas, January 17, 1997.                                   **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/

    local kern,aid1,i1,a;

    /*****************************
     ** Compute polynomial part **
     *****************************/
    aid1=p_air(seqa(0,1,rows(alpha)),p+1,cols(x));
    kern=0; i1=1; do while i1<=rows(alpha);
        a=alpha[i1];
        if a/=0;
            kern=kern+a*prodc((((x).^aid1[i1,.]))');
        endif;
    i1=i1+1; endo;
    kern=2*ln(abs(kern));

    /**********************************
     ** Compute central density part **
     **********************************/
    if nu>200;
        kern=kern - 0.5*ln(2*PI) - 0.5*sumc((x.*x)');
    else;
        kern=kern - cols(x)*ln(nu*PI)/2 + lnfact((nu+cols(x))/2-1)
             - lnfact(nu/2-1) - 0.5*(nu+cols(x))*ln(1+sumc((x.*x)')/nu);
    endif;

    retp(kern);
endp;






proc comp_logdens(y,x,b,R,alpha,p,nu);
/******************************************************************************
 ******************************************************************************
 **                                                                          **
 ** This routine computes ln() of the SNP density for the error process of   **
 ** the regression model y=x*b+e.                                            **
 **                                                                          **
 ** INPUT                                                                    **
 ** y    : (T x n1) matrix of dependent variables                            **
 ** x    : (T x n2) matrix of independent variables                          **
 ** b    : (n2 x n1) matrix of regression coefficients                       **
 ** R    : upper triangular matrix such that R'R=invpd(V), with V the        **
 **        (estimated) scaling matrix of the error process (y-x*b)           **
 ** alpha: the coeficients of the polynomial in the SNP expansion around the **
 **        Student t. alpha is parameterized as follows. Let A=alpha[i+1],   **
 **        then A is the coefficient of x[1]^k[1]*...*x[n]^k[n], where       **
 **        k[1]...k[n] is the n-digit representation of i in a system with   **
 **        base p.                                                           **
 ** p    : the degree of the polynomial in the Gallant-Nytchka               **
 **        representation. Note that this implies a restricted polynomial    **
 **        of degree p^n.                                                    **
 ** nu   : the degrees of freedom of the Student t distribution, the         **
 **        central part of the SNP expansion. nu>=200 gives the normal       **
 **        distribution. Note that nu must be large enough for the           **
 **        appropriate moments to exist due to the polynomial tails of       **
 **        the Student t and the polynomial in front of the central          **
 **        density in the SNP expansion.                                     **
 **                                                                          **
 ** OUTPUT                                                                   **
 ** kern : the ln(kernel) of the SNP density.                                **
 **                                                                          **
 ** Author: Andre Lucas, January 17, 1997.                                   **
 **                                                                          **
 ******************************************************************************
 ******************************************************************************/

    local mu,C0;

    /*******************************
     ** Compute the scaled errors **
     *******************************/
    if cols(x)/=0; y=(y-x*b)*R'; else; y=y*R'; endif;

    /**********************************************************************
     ** Compute the ln() density = ln(kernel) - ln(normalizing constant) **
     **********************************************************************/
    C0=comp_C(alpha,p,cols(y),nu,0);        @ compute normalizing constant @
    mu=comp_C(alpha,p,cols(y),nu,1)/C0;     @ compute mean and restrict to 0 @
    y = comp_logkernel(y+mu',alpha,p,nu) - ln(C0) + ln(prodc(diag(R)));
    retp(y);
endp;







proc comp_loglik(pars,data);
/*****************************************************************************
 *****************************************************************************
 **                                                                         **
 ** This routine computes the log-likelihood of the observations.           **
 **                                                                         **
 ** REQUIRED EXTERNAL VARIABLES:                                            **
 ** _glbr   : the postulated cointegration rank                             **
 ** _glbk   : the dimension of the process                                  **
 ** _glb_lownu  : lower bound for the degrees of freedom parameter          **
 ** _glbp_pol   : degree of the polynomial in the Gallant-Nytchka expansion **
 ** _glb_usevar : a vector of the same lenght as the __FULL__ parameter     **
 **               vector. The non-zero elements in _glb_usevar indicate     **
 **               that a parameter is used in the optimization. This device **
 **               is needed for restricting, e.g., the high polynomial      **
 **               terms in the Gallant-Nytchka representation to zero.      **
 **                                                                         **
 ** INPUT:                                                                  **
 ** pars : (sub)vector of full parameter vector.                            **
 ** data : (T x (n1+n2)) matrix. The first n1 columns contain the dependent **
 **        variables, the remaining columns the regressors.                 **
 **                                                                         **
 ** OUTPUT: (T x _glbk) matrix with the log likelihood of the observations  **
 **                                                                         **
 ** Author: Andre Lucas, January 17, 1997                                   **
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/

    local aid,y,x,a,b,g,v,alpha,nu,a22,R;


    /******************************************
     ** Split data matrix in regression form **
     ******************************************/
    y=data[.,1:_glbk]; x=data[.,_glbk+1:cols(data)];

    /*******************************************************************
     ** Construct the FULL parameter vector from pars and _glb_usevar **
     *******************************************************************/
    _glb_usevar=(_glb_usevar./=0);
    if sumc(_glb_usevar)/=rows(pars);
        "*** NUMBER OF PARAMETERS PASSED TO COMP_LOGLIK() INCORRECT ***";
        "    PASSED ROWS(PARS)  = " rows(pars);
        "    NEEDED _GLB_USEVAR = " _glb_usevar;
        stop;
    endif;
    pars=pars|0;
    aid=cumsumc(_glb_usevar).*_glb_usevar + rows(pars)*(1-_glb_usevar);
    pars=pars[aid];

    /**************************************************
     ** Get the parameters from the parameter vector **
     ** and reshape in the appropriate forms.        **
     **************************************************/
    {a,b,g,v,alpha,nu,a22}=getpars(pars,_glbk,_glbr,cols(x)-_glbk);
    if _glbr==0; b=zeros(_glbk,_glbk); else; b=a*b'; endif;
    if _glbr<_glbk;
        b[_glbr+1:_glbk,_glbr+1:_glbk]=b[_glbr+1:_glbk,_glbr+1:_glbk]+a22;
    endif;
    if ismiss(g)/=1; b=b~g; endif;
    R=chol(invpd(v));
    if scalerr(R)==20; retp(miss(1,1)+y[.,1]); endif;

    retp( comp_logdens(y,x,b',R,alpha,_glbp_pol,nu) );
endp;







proc (7)=getpars(pars,k,r,rem);
/***************************************************************************
 ***************************************************************************
 **                                                                       **
 ** {A,B,G,V,alpha,nu,a22} = getpars(pars,k,r,rem);                       **
 **                                                                       **
 ** This routine forms the appropriate parameters from the full parameter **
 ** vector. The parameter vector is assumed to be built up as follows.    **
 ** #Elements  Parameter                                                  **
 ** 1          sqrt(nu-_glb_lownu)                                        **
 ** k(k+1)/2   nonzero elements of vecr(chol(invpd(scaling matrix)))      **
 ** k*r        vecr(A), with A the error correction parameters            **
 ** (k-r)r     vecr(b), with (I b')' the cointegrating vectors            **
 ** (k-r)^2    vecr(a22), the unit root parameters                        **
 ** k*rem      vecr(G), remaining regression parameters                   **
 ** ??         alpha (first element restricted to unity)                  **
 **                                                                       **
 ** REQUIRED EXTERNAL VARIABLES:                                          **
 ** _glb_lownu : lower bound for the degrees of freedom parameter         **
 **                                                                       **
 ** INPUT:                                                                **
 ** pars  : parameter vector, built up as described above.                **
 ** k     : dimension of the stochastic process                           **
 ** r     : postulated cointegrating rank                                 **
 ** rem   : number of regressors seperate from (linear combinations) of   **
 **         lagged levels/cointegrating relations.                        **
 **                                                                       **
 ** OUPUT:                                                                **
 ** A     : (k x r) matrix containing error correction parameters.        **
 **         Missing value (scalar) if such parameters are absent.         **
 ** B     : (k x r) matrix of cointegrating vectors.                      **
 **         Missing value (scalar) if such parameters are absent.         **
 ** G     : (k x n1) matrix containing the coefficients of the remaining  **
 **         regressors. Missing value (scalar) if such parameters are     **
 **         absent.                                                       **
 ** V     : (k x k) matrix, scaling matrix of the errors (covariance).    **
 ** alpha : parameters containing the coefficients of the polynomial in   **
 **         the SNP expansion of the density.                             **
 ** nu    : degrees of freedom parameter, parameterized as _glb_lownu +   **
 **         par^2, with the par the maxlik parameter.                     **
 ** a22   : ((k-r) x (k-r)) matrix containing the unit root parameters.   **
 **                                                                       **
 ** Author: Andre Lucas, January 17, 1997                                 **
 **                                                                       **
 ***************************************************************************
 ***************************************************************************/

    local A,B,G,V,aid1,aid2,had,alpha,a22,nu;

    /******************
     ** construct nu **
     ******************/
    nu=_glb_lownu+pars[1]*pars[1]; pars=pars[2:rows(pars)];

    /*****************
     ** construct V **
     *****************/
    aid1=shiftr(ones(k,k),seqa(0,1,k),0);
    aid2=pars[1:((k+1)*k)/2]|0; A=vecr(aid1); B=(A.==0);
    A=cumsumc(A).*(1-B) + rows(aid2)*B; aid2=aid2[A];
    aid2=reshape(aid2,k,k); V=invpd(aid2'aid2);
    pars=pars[((k+1)*k)/2+1:rows(pars)];

    /*****************
     ** construct A **
     *****************/
    if r > 0;
        A=pars[1:r*k]; A=reshape(A,k,r);
        pars=pars[r*k+1:rows(pars)];
    else;
        A=miss(1,1);
    endif;

    /*****************
     ** construct B **
     *****************/
    if r > 0;
        if r==k;
            B=eye(k);
        else;
            B=pars[1:r*(k-r)]; B=eye(r)|reshape(B,k-r,r);
            pars=pars[r*(k-r)+1:rows(pars)];
        endif;
    else; B=miss(1,1); endif;

    /*******************
     ** construct a22 **
     *******************/
    if k-r > 0;
        a22=pars[1:(k-r)*(k-r)]; a22=reshape(a22,k-r,k-r);
        if (k-r)*(k-r)<rows(pars);
            pars=pars[(k-r)*(k-r)+1:rows(pars)];
        else;
            pars={};
        endif;
    else; a22=miss(1,1); endif;

    /*****************
     ** construct G **
     *****************/
    if rem > 0;
        G=pars[1:rem*k]; G=reshape(G,k,rem);
        if k*rem < rows(pars);
            pars=pars[k*rem+1:rows(pars)];
        else;
            pars={};
        endif;
    else;
        G=miss(1,1);
    endif;

    /*********************
     ** construct alpha **
     *********************/
    alpha=1|pars;

    retp(A,B,G,V,alpha,nu,a22);
endp;







proc (1)=putpars(A,B,G,V,alpha,nu,a22);
/***************************************************************************
 ***************************************************************************
 **                                                                       **
 **             {pars}=putpars(A,B,G,V,alpha,nu,a22);                     **
 **                                                                       **
 ** This routine forms the appropriate full parameter vector from the     **
 ** separate parameter matrices. The parameter vector is assumed to be    **
 ** built up as follows.                                                  **
 ** #Elements  Parameter                                                  **
 ** 1          sqrt(nu-_glb_lownu)                                        **
 ** k(k+1)/2   nonzero elements of vecr(chol(invpd(V=scaling matrix)))    **
 ** k*r        vecr(A), with A the error correction parameters            **
 ** (k-r)r     vecr(b), with (I b')' the cointegrating vectors            **
 ** (k-r)^2    vecr(a22), the unit root parameters                        **
 ** k*rem      vecr(G), remaining regression parameters                   **
 ** ??         alpha (first element restricted to unity)                  **
 **                                                                       **
 ** REQUIRED EXTERNAL VARIABLES:                                          **
 ** _glb_lownu : lower bound for the degrees of freedom parameter         **
 **                                                                       **
 ** INPUT:                                                                **
 ** A     : (k x r) matrix containing error correction parameters.        **
 **         Missing value (scalar) if such parameters are absent.         **
 ** B     : (k x r) matrix of cointegrating vectors.                      **
 **         Missing value (scalar) if such parameters are absent.         **
 ** G     : (k x n1) matrix containing the coefficients of the remaining  **
 **         regressors. Missing value (scalar) if such parameters are     **
 **         absent.                                                       **
 ** V     : (k x k) matrix, scaling matrix of the errors (covariance).    **
 ** alpha : parameters containing the coefficients of the polynomial in   **
 **         the SNP expansion of the density.                             **
 ** nu    : degrees of freedom parameter, parameterized as _glb_lownu +   **
 **         par^2, with the par the maxlik parameter.                     **
 ** a22   : ((k-r) x (k-r)) matrix containing the unit root parameters.   **
 **                                                                       **
 ** OUPUT:                                                                **
 ** pars  : parameter vector, built up as described above.                **
 **                                                                       **
 ** Author: Andre Lucas, January 17, 1997                                 **
 **                                                                       **
 ***************************************************************************
 ***************************************************************************/

    local k,r,pars,aid;

    /******************************************
     ** set dimension and cointegrating rank **
     ******************************************/
    k=cols(V); if ismiss(A)==1; r=0; else; r=cols(B); endif;

    /************
     ** put nu **
     ************/
    pars=sqrt(nu-_glb_lownu);

    /***********
     ** put V **
     ***********/
    aid=vecr(chol(invpd(V)));
    pars=pars|selif(aid,vecr(shiftr(ones(k,k),seqa(0,1,k),0)));

    /***********
     ** put A **
     ***********/
    if r > 0;
        B=B'; A=A*B[1:r,1:r]; B=inv(B[1:r,1:r])*B; B=B';
        pars=pars|vecr(A);
    endif;

    /***********
     ** put B **
     ***********/
    if (r > 0) and (r /= k); pars=pars|vecr(B[r+1:k,.]); endif;

    /*************
     ** put a22 **
     *************/
    if (k-r > 0); pars=pars|vecr(a22); endif;

    /***********
     ** put G **
     ***********/
    if ismiss(G)/=1; pars=pars|vecr(G); endif;

    /***************
     ** put alpha **
     ***************/
    if rows(alpha)>1; pars=pars|alpha[2:rows(alpha)]; endif;

    retp(pars);
endp;







proc (2)=gen_xy(series,p,dets);
/*****************************************************************************
 *****************************************************************************
 ** This procedure returns a y matrix and an x matrix for regression.       **
 ** The y matrix contains the first differences of the series,              **
 ** the x matrix the lagged levels, then (p-1) lagged first differences,    **
 ** then the deterministic terms in dets.                                   **
 ** If dets has the same number of columns as series, the appropriate rows  **
 ** are horizontally appended to the x matrix. If dets is a nonnegative     **
 ** scalar, a deteriministic time polynomial is added to the x matrix. The  **
 ** order of the polynomial is floor(dets). In all other cases an error     **
 ** is printed.                                                             **
 **                                                                         **
 ** Date: 8 May 1995                                                        **
 ** Author: Andre Lucas                                                     **
 *****************************************************************************
 *****************************************************************************/
    local n,x,y,i1;

    if p<1;
        "*** ERROR: NUMBER OF LAGS MUST BE GREATER THAN ZERO ***";
        stop;
    endif;

    n=rows(series); _glbp=p;
    x=series[p:n-1,.]; y=series[p+1:n,.]-x;
    i1=1; do while i1<p;
        x=x~(series[p+1-i1:n-i1,.]-series[p-i1:n-1-i1,.]);
    i1=i1+1; endo;

    if (rows(dets)==cols(dets)) and (rows(dets)==1);
        dets=floor(dets);
        if (dets<0);
            "...WARNING: NO DETERMINISTICS ADDED";
        else;
            x=x~(seqa(1,1,rows(x)).^(seqa(0,1,dets+1)'));
        endif;
    elseif rows(dets)==n; x=x~dets[p+1:n,.];
    else;
        "*** ERROR: WRONG SPECIFICATION FOR DETERMINISTICS ***";
        stop;
    endif;

    retp(y,x);
endp;







proc (6)=Gauss_ml(y,x,coint);
/***************************************************************************
 ***************************************************************************
 ** Performs Gaussian ML for the model y(t)=x(t)'Gamma + e(t),            **
 ** with y(t) the t-th row of y. The first cols(y) columns of x must      **
 ** contain the lagged levels of the original series, while y contains    **
 ** the first differences of the original series.                         **
 **                                                                       **
 ** y: NxK1                                                               **
 ** x: NxK2 (K2>=K1)                                                      **
 ** coint: the cointegration rank that is fitted                          **
 **                                                                       **
 ** Output: {A,B,G,V,errors}                                              **
 **  V covariance matrix of the errors,                                   **
 **  A*B' parameters of the lagged levels,                                **
 **  G remaining regression parameters                                    **
 **                                                                       **
 ** Date: 8 May 1995                                                      **
 ** Author: Andre Lucas                                                   **
 ***************************************************************************
 ***************************************************************************/

    local n,k1,k2,xaid,ry,rx,m00,m10,m11,a,b,g,v,vvals,vecs;

    /**********************************
     ** check and initialize formats **
     **********************************/
    n=rows(y); k1=cols(y); k2=cols(x);
    if n/=rows(x); "*** ERROR: ROWS(Y)/=ROWS(X) ***"; stop; endif;
    if k2<k1; "*** ERROR: COLS(X)<COLS(Y) ***"; stop; endif;
    if (coint<0) or (coint>k1); "*** ERROR: WRONG COINTEGRATION RANK ***";
       stop; endif;

    /**************************************************************
     ** initialize regression matrix for preliminary regressions **
     **************************************************************/
    if k1==k2; xaid=miss(1,1);
    else; xaid=x[.,k1+1:k2]; endif;

    /*************************************
     ** perform preliminary regressions **
     *************************************/
    if ismiss(xaid)/=1;
        ry=y-xaid*(y/xaid); rx=x[.,1:k1]-xaid*(x[.,1:k1]/xaid);
    else;
        ry=y; rx=x;
    endif;

    /***********************************************
     ** construct matrices for eigenvalue problem **
     ***********************************************/
    m00=ry'ry; m10=rx'ry; m11=rx'rx;
    {vvals,vecs}=eighv(m11); vvals=diagrv(eye(k1),vvals.^(-0.5));
    vvals=vecs*vvals*vecs'; m00=vvals*m10*invpd(m00)*m10'*vvals;
    {m00,m10}=eighv(m00); m00=rev(sortc(m00~(m10'),1)); m11=m00[.,1];
    m10=m00[.,2:k1+1]'; m10=vvals*m10;

    /*****************************
     ** construct estimate of B **
     *****************************/
    if coint>0; b=m10[.,1:coint];
    else; b=miss(1,1); endif;

    /*****************************
     ** construct estimate of A **
     *****************************/
    if coint>0; a=(ry/(rx*b))';
    else; a=miss(1,1); endif;

    /*****************************
     ** construct estimate of V **
     *****************************/
    if coint>0; v=ry-rx*b*a'; m00=v; v=v'v/rows(v);
    else; v=ry'ry/rows(ry); m00=ry; endif;

    /***********************************************************
     ** construct estimate of remaining regression parameters **
     ***********************************************************/
    if k2>k1;
        if coint>0; ry=y-x[.,1:k1]*b*a';
        else; ry=y; endif;
        g=(ry/xaid)';
    else; g=miss(1,1); endif;

    /********************
     ** return results **
     ********************/
    if coint>0;
        xaid=b[1:coint,1:coint];
        a=a*xaid'; b=b*inv(xaid);
    endif;
    if coint==cols(y); _glb_johansen=-1;
    else; _glb_johansen=-rows(y)*sumc(ln(1-m11[coint+1:rows(m11)])); endif;
    retp(a,b,g,v,m00,m11);  @ m00=residuals, m11=eigenvalues @
endp;



proc (1)=comp_strt(y,x,with_a22,par,snp_pol);
/*****************************************************************************
 *****************************************************************************
 ** This routine computes starting values for the optimization.             **
 **                                                                         **
 ** EXTERNAL VARIABLES NEEDED IN THE ROUTINE:                               **
 ** _glbr   : the postulated cointegration rank                             **
 ** _glbk   : the dimension of the process                                  **
 ** _glb_lownu  : lower bound for the degrees of freedom parameter          **
 ** _glbp_pol   : degree of the polynomial in the Gallant-Nytchka expansion **
 ** _glb_mett   : if /= 0, degrees of freedom also estimated                **
 ** _glb_usevar : a vector of the same lenght as the __FULL__ parameter     **
 **               vector. The non-zero elements in _glb_usevar indicate     **
 **               that a parameter is used in the optimization. This device **
 **               is needed for restricting, e.g., the high polynomial      **
 **               terms in the Gallant-Nytchka representation to zero.      **
 **               _glb_usevar is set in this routine.                       **
 ** _glb_true_degree : if 1, then restrictions on alpha are imposed in      **
 **                    order to obtain a full _glbp_pol-th order polynomial **
 **                    instead of a restricted _glbp_pol*_glbk-th order     **
 **                    polynomial.                                          **
 **                                                                         **
 **                                                                         **
 ** INPUT:                                                                  **
 ** y : T x n1 matrix of dependent variables                                **
 ** x : T x n2 matrix of independent variables                              **
 ** with_a22 : scalar; if 1 then incorporate a22 in the set of parameters   **
 **            to optimize over                                             **
 ** par      : parameter vector containing estimates for SNP approximation  **
 **            snp_pol                                                      **
 ** snp_pol  : scalar, degree of the SNP approximation used to obtain the   **
 **            starting values in par; par is only used if snp_pol <=       **
 **            _glbp_pol                                                    **
 **                                                                         **
 ** OUTPUT:                                                                 **
 ** strt : starting value of parameters                                     **
 **                                                                         **
 ** Author: Andre Lucas, January 17, 1997                                   **
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/

    local A,B,G,V,e,ev,nu,alpha,a22;

    if (snp_pol <= _glbp_pol) and (rows(par) > 0);
        par=par|0;
        e=cumsumc(_glb_usevar).*_glb_usevar + rows(par)*(1-_glb_usevar);
        par=par[e];
        {A,B,G,V,alpha,nu,a22}=getpars(par,_glbk,_glbr,cols(x)-cols(y));
        e=p_air(seqa(0,1,(1+snp_pol)^_glbk),snp_pol+1,_glbk)';
        e=sumc(e.*((1+_glbp_pol)^rev(seqa(0,1,_glbk)))); ev=alpha;
        alpha=zeros((1+_glbp_pol)^_glbk,1); alpha=rndu(rows(alpha),1)*0.1;
        alpha[1]=1; alpha[e+1]=ev;
    else;
        {A,B,G,V,e,ev}=Gauss_ml(y,x,_glbr);
        nu=_glb_lownu+rndu(1,1)*(_glb_mett/=0)/10;
        alpha=zeros((1+_glbp_pol)^_glbk,1); alpha[1]=1;
        if _glbk-_glbr > 0;
            a22=zeros(_glbk-_glbr,_glbk-_glbr)+(with_a22==1);
        else;
            a22=miss(1,1);
        endif;
    endif;

    /*****************
     ** normalize B **
     *****************/
    e=putpars(A,B,G,V,alpha,nu,a22); strt=e;
    {A,B,G,V,alpha,nu,a22}=getpars(e,_glbk,_glbr,cols(x)-cols(y));

    /***************************************
     ** select parameters in optimization **
     ***************************************/
    e=(B./=0) .and (B./=1); e=e+B.*(1-e);
    ev=shiftr(ones(_glbk,_glbk),seqa(0,1,_glbk),0); ev=invpd(ev'ev);
    nu=p_air(seqa(0,1,(1+_glbp_pol)^_glbk),_glbp_pol+1,_glbk)';
    nu=sumc(nu); nu=(nu .<= _glbp_pol + (_glb_true_degree/=1)*maxc(nu));
    _glb_usevar=putpars(0*A+1,e,0*G+1,ev,nu,_glb_lownu+(_glb_mett./=0),a22);

    retp(selif(strt,_glb_usevar));
endp;





proc (2)=comp_psi_p(par,data);
/*****************************************************************************
 *****************************************************************************
 **                                                                         **
 **              proc {psi,psip}=comp_psi_p(par,data);                      **
 **                                                                         **
 ** This routine computes the 1st and 2nd order derivatives of the SNP      **
 ** expansion.                                                              **
 **                                                                         **
 ** REQUIRED EXTERNAL VARIABLES:                                            **
 ** _glb_usevar : a vector of the same lenght as the __FULL__ parameter     **
 **               vector. The non-zero elements in _glb_usevar indicate     **
 **               that a parameter is used in the optimization. This device **
 **               is needed for restricting, e.g., the high polynomial      **
 **               terms in the Gallant-Nytchka representation to zero.      **
 ** __output    : control of output level as in maxlik()                    **
 ** Further all external variables for comp_loglik() are required.          **
 **                                                                         **
 ** INPUT:                                                                  **
 ** par  : (sub)vector of full parameter vector.                            **
 ** data : (T x (n1+n2)) matrix. The first n1 columns contain the dependent **
 **        variables, the remaining columns the regressors.                 **
 **                                                                         **
 ** OUTPUT:                                                                 **
 ** psi  : (T x k) matrix of unscaled pseudo scorres                        **
 ** psip : (k x k) matrix of (mean) Hessian (= pseudo information)          **
 **                                                                         **
 ** Author: Andre Lucas, February 10, 1997                                  **
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/

    local aid,y,x,a,b,g,v,alpha,nu,a22,p,R,eps,dy,absy,dh,hes,hesmin,
          i1,i2,pars;

    /******************************************
     ** Split data matrix in regression form **
     ******************************************/
    y=data[.,1:_glbk]; x=data[.,_glbk+1:cols(data)]; p=_glbp_pol;

    /*******************************************************************
     ** Construct the FULL parameter vector from pars and _glb_usevar **
     *******************************************************************/
    _glb_usevar=(_glb_usevar./=0);
    if sumc(_glb_usevar)/=rows(par);
        "*** NUMBER OF PARAMETERS PASSED TO COMP_LOGLIK() INCORRECT ***";
        "    PASSED ROWS(PARS)  = " rows(par);
        "    NEEDED _GLB_USEVAR = " _glb_usevar;
        stop;
    endif;
    par=par|0;
    aid=cumsumc(_glb_usevar).*_glb_usevar + rows(par)*(1-_glb_usevar);
    par=par[aid];

    /*******************************************
     ** if nu is not optimized over,          **
     ** fix it at _glb_lownu instead of zero. **
     *******************************************/
    if _glb_usevar[1]==0; par[1]=0; endif;

    /**************************************************
     ** Get the parameters from the parameter vector **
     ** and reshape in the appropriate forms.        **
     **************************************************/
    {a,b,g,v,alpha,nu,a22}=getpars(par,_glbk,_glbr,cols(x)-_glbk);
    if _glbr==0; b=zeros(_glbk,_glbk); else; b=a*b'; endif;
    if _glbr<_glbk;
        b[_glbr+1:_glbk,_glbr+1:_glbk]=b[_glbr+1:_glbk,_glbr+1:_glbk]+a22;
    endif;
    if ismiss(g)/=1; b=b~g; endif;
    R=chol(invpd(v)); y=(y-x*b');

    /********************************
     ** Compute Delta for gradient **
     ********************************/
    eps=g[.,cols(g)]; absy=abs(eps);
    if eps /= 0; dy=eps./absy; else; dy=1; endif;
    dh=(absy.<1E-2); eps=6.0554544523933429E-6;
    dh=eps*(1E-2*dh + absy.*(1-dh)).*dy;
    pars=par; par=selif(pars,_glb_usevar);

    /**********************
     ** Compute gradient **
     ** and Hessian      **
     **********************/
    hes=zeros(cols(y),cols(y)); hesmin=hes; absy=y;
    i1=1; do while i1<=cols(y);
        data[.,i1] = data[.,i1]-dh[i1];
        absy[.,i1] = comp_loglik(par,data);

        i2=i1; do while i2<=cols(y);
            data[.,i2] = data[.,i2]-dh[i2];
            aid = comp_loglik(par,data);
            data[.,i2] = data[.,i2]+2*dh[i2];
            aid = (aid-comp_loglik(par,data));
            data[.,i2] = data[.,i2]-dh[i2];
            hes[i1,i2] = meanc(aid)/(4*dh[i2]*dh[i1]);
        i2=i2+1; endo;

        data[.,i1] = data[.,i1]+2*dh[i1];
        absy[.,i1] = (absy[.,i1]-comp_loglik(par,data))./(2*dh[i1]);

        i2=i1; do while i2<=cols(y);
            data[.,i2] = data[.,i2]-dh[i2];
            aid = comp_loglik(par,data);
            data[.,i2] = data[.,i2]+2*dh[i2];
            aid = (aid-comp_loglik(par,data));
            data[.,i2] = data[.,i2]-dh[i2];
            hesmin[i1,i2]=meanc(aid)/(4*dh[i2]*dh[i1]);
        i2=i2+1; endo;

        data[.,i1] = data[.,i1]-dh[i1];
    i1=i1+1; endo;
    hes=hes-hesmin; hes=hes+lowmat(diagrv(hes',zeros(_glbk,1)));

    retp(absy,hes);
endp;






proc (4)=comp_grheop(par,data,k,r);
/*****************************************************************************
 *****************************************************************************
 **                                                                         **
 **              proc {gr,hes,op,lm}=comp_grheop(par,data,k,r);             **
 **                                                                         **
 ** This routine computes the gradient of the log-likelihood, the outer     **
 ** product of gradients, the Hessian, and the LM test for cointegration.   **
 **                                                                         **
 ** REQUIRED EXTERNAL VARIABLES:                                            **
 ** _glb_usevar : a vector of the same lenght as the __FULL__ parameter     **
 **               vector. The non-zero elements in _glb_usevar indicate     **
 **               that a parameter is used in the optimization. This device **
 **               is needed for restricting, e.g., the high polynomial      **
 **               terms in the Gallant-Nytchka representation to zero.      **
 ** __output    : control of output level as in maxlik()                    **
 ** Further all external variables for comp_loglik() are required.          **
 **                                                                         **
 ** INPUT:                                                                  **
 ** par : (sub)vector of full parameter vector.                             **
 ** data: (T x (n1+n2)) matrix. The first n1 columns contain the dependent  **
 **        variables, the remaining columns the regressors.                 **
 ** k   : dimension of the stochastic system                                **
 ** r   : postulated cointegrating rank                                     **
 **                                                                         **
 ** OUTPUT:                                                                 **
 ** gr  : gradient of the log-likelihood (divided by T)                     **
 ** hes : hessian of the log-likelihood (divided by T)                      **
 ** op  : outer product of gradients (divided by T)                         **
 ** lm  : the LM test for cointegration                                     **
 **                                                                         **
 ** Author: Andre Lucas, January 20, 1997                                   **
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/

    local gr,hes,opg,hese,opge,lm,i1,x,a,b,g,v,alpha,nu,a22;

    /**********************************
     ** Compute gradient and hessian **
     **********************************/
    {gr,hese}=comp_psi_p(par,data); opge=gr'gr/rows(gr);

    /***********************************
     ** Compute full parameter vector **
     ***********************************/
    par=par|0; i1=cumsumc(_glb_usevar).*_glb_usevar + rows(par)*(1-_glb_usevar);
    par=par[i1];

    /******************************************************
     ** Build derivative of epsilon (=regression errors) **
     ** with respect to regression parameters            **
     ******************************************************/
    x={}; {a,b,g,v,alpha,nu,a22}=getpars(par,k,r,cols(data)-2*k); a22={};
    if cols(data)>2*k;
        x=data[.,2*k+1:cols(data)].*.eye(k);
        a22=0*x[1:k,1:cols(x)/k];
    endif;
    if r > 0;
        x=((data[.,k+1:2*k]*b).*.eye(k))~x;
        a22=0*x[1:k,1:cols(x)/k];
    endif;
    a22=eye(k)~a22;
    if k > r;
        if r>0;
            a=a~(zeros(r,k-r)|eye(k-r));
            a22=a22.*.(zeros(k-r,r)~eye(k-r));
        else;
            a=eye(k-r);
            a22=a22.*.a;
        endif;
        x=(data[.,k+r+1:2*k].*.a)~x;
    endif;
    x=-x;

    /***********************************
     ** Compute gradient, hessian and **
     ** outer product of gradients    **
     ***********************************/
    gr=sumc(vecr(gr).*x)/rows(data); hes=0; opg=0;
    if __output==2; locate 1,1; blank; endif;
    if __output==2; locate 1,1; "Observation "; endif;
    i1=0; do while i1<rows(data);
        hes=hes+x[i1*k+1:i1*k+k,.]'*hese*x[i1*k+1:i1*k+k,.];
        opg=opg+x[i1*k+1:i1*k+k,.]'*opge*x[i1*k+1:i1*k+k,.];
        if (i1%25)==0 and __output==2; locate 1,13; i1; endif;
    i1=i1+1; endo;
    hes=hes/rows(data); opg=opg/rows(data);

    /*********************
     ** Compute LM test **
     *********************/
    lm=a22*inv(hes)*gr; hese=a22*inv(hes)*opg*inv(hes)*a22';
    lm=rows(data)*lm'invpd(hese)*lm;

    if __output==2; locate 1,1; blank; endif;
    retp(gr,hes,opg,lm);

endp;




proc (2)=comp_corr(par,data);
/*****************************************************************************
 *****************************************************************************
 **                                                                         **
 ** {c1,c2} = comp_corr(par,data);                                          **
 **                                                                         **
 ** This routine computes the correlations, i.e., the nuisance parameters,  **
 ** that enter the limiting distribution of the LM test for cointegration   **
 ** based on tha SNP density estimate.                                      **
 **                                                                         **
 ** REQUIRED EXTERNAL VARIABLES:                                            **
 ** _glbr   : the postulated cointegration rank                             **
 ** _glbk   : the dimension of the process                                  **
 ** _glb_lownu  : lower bound for the degrees of freedom parameter          **
 ** _glbp_pol   : degree of the polynomial in the Gallant-Nytchka expansion **
 ** _glb_usevar : a vector of the same lenght as the __FULL__ parameter     **
 **               vector. The non-zero elements in _glb_usevar indicate     **
 **               that a parameter is used in the optimization. This device **
 **               is needed for restricting, e.g., the high polynomial      **
 **               terms in the Gallant-Nytchka representation to zero.      **
 **                                                                         **
 ** INPUT:                                                                  **
 ** par  : (sub)vector of full parameter vector.                            **
 ** data : (T x (n1+n2)) matrix. The first n1 columns contain the dependent **
 **        variables, the remaining columns the regressors.                 **
 **                                                                         **
 ** OUTPUT:                                                                 **
 ** c1   : correlations for A\lood\trans\mu \= 0                            **
 ** c2   : correlations for A\lood\trans\mu == 0                            **
 **                                                                         **
 ** Author: Andre Lucas, January 20, 1997                                   **
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/

    local aid,y,x,a,b,g,v,alpha,nu,a22,p,R,absy,hes,alood,blood,c1,c2,zeta1,mu;

    /**************************************
     ** Compute "gradient" and "hessian" **
     **************************************/
    {absy,hes}=comp_psi_p(par,data); absy=absy*inv(hes);

    /******************************************
     ** Split data matrix in regression form **
     ******************************************/
    y=data[.,1:_glbk]; x=data[.,_glbk+1:cols(data)]; p=_glbp_pol;

    /*******************************************************************
     ** Construct the FULL parameter vector from pars and _glb_usevar **
     *******************************************************************/
    _glb_usevar=(_glb_usevar./=0);
    if sumc(_glb_usevar)/=rows(par);
        "*** NUMBER OF PARAMETERS PASSED TO COMP_LOGLIK() INCORRECT ***";
        "    PASSED ROWS(PARS)  = " rows(par);
        "    NEEDED _GLB_USEVAR = " _glb_usevar;
        stop;
    endif;
    par=par|0;
    aid=cumsumc(_glb_usevar).*_glb_usevar + rows(par)*(1-_glb_usevar);
    par=par[aid];

    /*******************************************
     ** if nu is not optimized over,          **
     ** fix it at _glb_lownu instead of zero. **
     *******************************************/
    if _glb_usevar[1]==0; par[1]=0; endif;

    /**************************************************
     ** Get the parameters from the parameter vector **
     ** and reshape in the appropriate forms.        **
     **************************************************/
    {a,b,g,v,alpha,nu,a22}=getpars(par,_glbk,_glbr,cols(x)-_glbk);
    if _glbr==0; b=zeros(_glbk,_glbk); else; b=a*b'; endif;
    if _glbr<_glbk;
        b[_glbr+1:_glbk,_glbr+1:_glbk]=b[_glbr+1:_glbk,_glbr+1:_glbk]+a22;
    endif;
    if ismiss(g)/=1; b=b~g; endif;
    R=chol(invpd(v)); y=(y-x*b');

    /********************************************
     ** Check whether anything can be computed **
     ** and initialize Alood                   **
     ********************************************/
    if _glbk == _glbr; retp(miss(1,1),miss(1,1)); endif;
    if ismiss(a)==1; alood=eye(_glbk); else; alood=null(a'); endif;
    if ismiss(a)==1; blood=eye(_glbk); else; blood=null(b'); endif;

    /**************************
     ** Compute correlations **
     **************************/
    mu=g[.,cols(g)]; y=y*alood; absy=absy*alood; c1={};
    {c2,zeta1}=eighv(y'y); c2=invpd(zeta1*(sqrt(c2).*zeta1'));
    c2=sqrt(svd(c2*(y'absy)*invpd(absy'absy)*(absy'y)*c2'));
    if _glb_detpol>=0;
        zeta1=alood'mu; zeta1=null(zeta1');
        y=y*zeta1; {c1,zeta1}=eighv(y'y);
        c1=invpd(zeta1*(sqrt(c1).*zeta1'));
        c1=sqrt(svd(c1*(y'absy)*invpd(absy'absy)*(absy'y)*c1'));
    endif;

    retp(c1,c2);
endp;





proc (1)=sim_pval(n,test,c1,c2);
/*****************************************************************************
 *****************************************************************************
 **                                                                         **
 ** This routine simulates asymptotic p-values of the LM test for           **
 ** cointegration given estimates of the nuisance parameters.               **
 **                                                                         **
 ** REQUIRED EXTERNAL VARIABLES:                                            **
 ** _glbr   : the postulated cointegration rank                             **
 ** _glbk   : the dimension of the process                                  **
 **                                                                         **
 ** INPUT:                                                                  **
 ** n    : number of observations in the data                               **
 ** test : value of the test statistic                                      **
 ** c1   : correlations for A\lood\trans\mu \= 0                            **
 ** c2   : correlations for A\lood\trans\mu == 0                            **
 **                                                                         **
 ** OUTPUT: (3 x 1) vector containing the simulated p-values. 1st element   **
 ** for demeaned Brownian motion with drift, 2nd element for demeaned       **
 ** Brownian motion without drift, 3rd element for detrended Brownian       **
 ** motion without quadratic drift.                                         **
 **                                                                         **
 ** Author: Andre Lucas, January 20, 1997                                   **
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/

    local k,i1,i2,result,y,dy,d3,maxit,sel1,sel2,aid1,aid2;

    /****************
     ** Initialize **
     ****************/
    if ismiss(test)==1; retp(test); endif;
    n=minc(250|n); k=_glbk-_glbr; maxit=_glb_psim; i2=floor(maxit/10);
    result=zeros(maxit,3); sel2=eye(k+1); c1=0|c1;
    sel1=delif(sel2,seqa(1,1,k+1).==2);
    sel2=delif(sel2,seqa(1,1,k+1).==1);

    /*************************
     ** Perform simulations **
     *************************/
    if __output==2; locate 1,1; blank; endif;
    i1=1; do while i1<=maxit;
        if (i1%i2==0) and (__output>=2);
            locate 1,1; 100*i1/maxit "%";
        endif;
        dy=rndn(n,k)/sqrt(n); d3=rndn(n,k)/sqrt(n);
        y=zeros(1,k)|cumsumc(dy[1:n-1,.]);
        y=seqa(1/n,1/n,n)~y; y=y-meanc(y)';
        dy=y'dy; d3=y'd3; y=y'y/n;
        aid1=sel1'invpd(sel1*y*sel1')*sel1;
        aid2=(dy.*c1') + (d3.*sqrt(1-(c1^2)'));
        result[i1,1]=sumc(diag(aid2'aid1*aid2));
        aid1=sel2'invpd(sel2*y*sel2')*sel2;
        aid2=(dy.*c2') + (d3.*sqrt(1-(c2^2)'));
        result[i1,2]=sumc(diag(aid2'aid1*aid2));
        aid1=invpd(y); aid2=sel2*aid1*aid2;
        result[i1,3]=sumc(diag(aid2'invpd(sel2*aid1*sel2')*aid2));
    i1=i1+1; endo;
    if __output==2; locate 1,1; blank; endif;

    retp(meanc(result.>test));
endp;


proc (1)=polynorm(x);
/***************************
 ** draw from             **
 ** f(x) = pdfn(x)*x^6/15 **
 ***************************/
    retp(cdfn(x)-pdfn(x).*x.*(15+x.*x.*(5+x.*x))/15-num_el);   @ x^6 @
    retp((cdfn(x)-pdfn(x).*x.*(3+x.*x)/3-num_el).^2);          @ x^4 @
    retp((cdfn(x)-pdfn(x).*x-num_el).^2);                      @ x^2 @
endp;



